class PoorManPlatformerPhysics
  attr_accessor :grid, :inputs, :state, :outputs

  def tick
    defaults
    render
    calc
    process_inputs
  end

  def defaults
    state.tile_size               = 64
    state.gravity                 = -0.2
    state.previous_tile_size    ||= state.tile_size
    state.x                     ||= 0
    state.y                     ||= 800
    state.dy                    ||= 0
    state.dx                    ||= 0
    state.world                 ||= []
    state.world_lookup          ||= {}
    state.world_collision_rects ||= []
  end

  def render
    outputs.solids << grid.rect

    outputs.borders << state.world.map do |x, y|
      [x * state.tile_size,
       y * state.tile_size,
       state.tile_size,
       state.tile_size, 255, 255, 255]
    end

    outputs.borders << state.world_collision_rects.map do |e|
      [
        [e[:top],                             0, 170,   0],
        [e[:bottom],                          0, 100, 170],
        [e[:left_right],                    170,   0,   0],
      ]
    end

    outputs.borders << [state.x,
                        state.y,
                        state.tile_size,
                        state.tile_size,  0, 180, 0]
  end

  def calc_world_lookup
    if state.tile_size != state.previous_tile_size
      state.previous_tile_size = state.tile_size
      state.world_lookup = {}
    end

    return if state.world_lookup.keys.length > 0
    return unless state.world.length > 0

    state.world_lookup = {}
    state.world.each { |x, y| state.world_lookup[[x, y]] = true }
    state.world_collision_rects =
      state.world_lookup
          .keys
          .map do |coord_x, coord_y|
            s = state.tile_size
            x = s * coord_x
            y = s * coord_y
            {
              args:       [coord_x, coord_y],
              left_right: [x,     y + 4, s,     s - 6],
              top:        [x + 4, y + 6, s - 8, s - 6],
              bottom:     [x + 1, y - 1, s - 2, s - 8],
            }
          end
  end

  def calc_player
    state.dy += state.gravity

    calc_box_collision
    calc_edge_collision

    state.y += state.dy
    state.x += state.dx
    state.dx *= 0.8
  end

  def calc_edge_collision
    if state.y < 0
      state.y = 0
      state.dy = 0
    elsif state.y > 720 - state.tile_size
      state.y = 720 - state.tile_size
      state.dy = 0
    end

    if state.x >= 1280 - state.tile_size && state.dx > 0
      state.x = 1280 - state.tile_size
      state.dx = 0
    elsif state.x <= 0 && state.dx < 0
      state.x = 0
      state.dx = 0
    end
  end

  def collision_tollerance
    0.0
  end

  def collision_ceiling!
    return unless state.dy > 0

    player_rect = [state.x, state.y + 0.1, state.tile_size, state.tile_size]

    ceil_collisions = state.world_collision_rects
                          .find_all { |r| r[:bottom].intersects_rect?(player_rect, collision_tollerance) }
                          .first

    return unless ceil_collisions
    state.y = ceil_collisions[:bottom].y - state.tile_size
    state.dy = 0
  end

  def collision_left!
    return unless state.dx < 0
    player_rect = [state.x - 0.1, state.y, state.tile_size, state.tile_size]
    left_side_collisions = state.world_collision_rects
                               .find_all { |r| r[:left_right].intersects_rect?(player_rect, collision_tollerance) }
                               .first

    return unless left_side_collisions
    state.x = left_side_collisions[:left_right].right
    state.dx = 0
  end

  def collision_right!
    return unless state.dx > 0

    player_rect = [state.x + 0.1, state.y, state.tile_size, state.tile_size]
    right_side_collisions = state.world_collision_rects
                                .find_all { |r| r[:left_right].intersects_rect?(player_rect, collision_tollerance) }
                                .first

    return unless right_side_collisions
    state.x = right_side_collisions[:left_right].left - state.tile_size
    state.dx = 0
  end

  def collision_floor!
    return unless state.dy <= 0

    player_rect = [state.x, state.y - 0.1, state.tile_size, state.tile_size]

    floor_collisions = state.world_collision_rects
                           .find_all { |r| r[:top].intersects_rect?(player_rect, collision_tollerance) }
                           .first

    return unless floor_collisions
    state.y = floor_collisions[:top].top
    state.dy = 0
  end

  def calc_box_collision
    return unless state.world_lookup.keys.length > 0
    collision_floor!
    collision_left!
    collision_right!
    collision_ceiling!
  end

  def calc
    calc_world_lookup
    calc_player
  end

  def to_coord point
    [point.x.idiv(state.tile_size), point.y.idiv(state.tile_size)]
  end

  def process_inputs
    if inputs.mouse.down
      state.world_lookup = {}
      x, y = to_coord inputs.mouse.down.point

      if state.world.any? { |loc| loc == [x, y] }
        state.world = state.world.reject { |loc| loc == [x, y] }
      else
        state.world << [x, y]
      end
    end

    if inputs.keyboard.key_up.right
      state.dx = 0
    elsif inputs.keyboard.key_up.left
      state.dx = 0
    end

    if inputs.keyboard.key_held.right
      state.dx =  3
    elsif inputs.keyboard.key_held.left
      state.dx = -3
    end

    if inputs.keyboard.key_held.space
      state.dy = 5
    end
  end
end

$platformer_physics = PoorManPlatformerPhysics.new

def tick args
  $platformer_physics.grid    = args.grid
  $platformer_physics.inputs  = args.inputs
  $platformer_physics.state    = args.state
  $platformer_physics.outputs = args.outputs
  $platformer_physics.tick
end
