class TextedBasedGame
  def initialize game_state_provider
    @game_state_provider = game_state_provider
  end

  def state
    @game_state_provider.state
  end

  def default
    state.engine_tick_count ||= 0
    state.active_module     ||= :room
    state.fire_progress     ||= 0
    state.fire_ready_in     ||= 10
    state.previous_fire     ||= :dead
    state.fire              ||= :dead
  end

  def room_title
    return "a room that is dark" if state.fire == :dead
    return "a room that is lit"
  end

  def tick
    default
    show_intro
    state.engine_tick_count += 1
    tick_fire
  end

  def tick_fire
    return if state.active_module == :alert
    state.fire_progress += 1
    state.fire_progress = state.fire_progress.lesser(state.fire_ready_in)
  end

  def show_intro
    return unless state.engine_tick_count == 0
    set_story_line "awake."
  end

  def set_story_line story_line
    state.story_line    = story_line
    state.active_module = :alert
  end

  def clear_storyline
    state.active_module = :none
    state.story_line = nil
  end

  def should_tick?
    state.tick_count.mod_zero?(60)
  end

  def active_module
    state.active_module
  end

  def story_line
    state.story_line
  end

  def go_to_room
    state.active_module = :room
  end

  def light_fire_progress
    state.fire_progress.fdiv(10)
  end

  def fire
    state.fire
  end

  def fire_ready?
    state.fire_progress == state.fire_ready_in
  end

  def light_fire
    return unless fire_ready?
    state.fire = :roaring
    state.fire_progress = 0
    if state.fire != state.previous_fire
      set_story_line "the fire is #{state.fire}."
      state.previous_fire = state.fire
    end
  end

  def save
    $gtk.save_state('game_state.txt')
  end

  def load
    success = $gtk.load_state('game_state.txt')
    return if success
    set_story_line "no game to load. press save first."
  end

  def reset
    $gtk.reset
  end
end

class TextedBasedGamePresenter
  attr_accessor :state, :outputs, :inputs

  def initialize
    @game = TextedBasedGame.new self
    @button_list ||= {}
  end

  def button id, x, y, text
    @button_list[id] ||= {
      id: id,
      text: text,
      primitives: [
        [x + 10, y + 30, text, 2, 0].label,
        [x, y, 300, 50].border,
      ]
    }

    @button_list[id][:primitives]
  end

  def progress_bar id, x, y, text, percentage
    @button_list[id] = {
      id: id,
      text: text,
      primitives: [
        [x, y, 300, 50, 100, 100, 100].solid,
        [x + 10, y + 30, text, 2, 0].label,
        [x, y, 300, 50].border,
      ]
    }

    @button_list[id][:primitives][0][2] = 300 * percentage
    @button_list[id][:primitives]
  end

  def render_alert
    return unless game.active_module == :alert

    outputs.labels << [640, 480, game.story_line, 5, 1]
    outputs.primitives << button(:alert_dismiss, 490, 380, "close")
  end

  def render
    outputs.labels << [10, 30, state.tick_count]
    render_alert
    render_room
    render_highlights
  end

  def render_highlights
    state.layout.highlights.each do |h|
      h.lifetime -= 1
    end

    outputs.solids << state.layout.highlights.map do |h|
      [h.x, h.y, h.w, h.h, h.color, 255 * h.lifetime / h.max_lifetime]
    end

    state.layout.highlights = state.layout.highlights.reject { |h| h.lifetime <= 0 }
  end

  def render_room
    return unless game.active_module == :room
    outputs.labels << [640, 700, game.room_title, 4, 1]
    outputs.primitives << progress_bar(:light_fire, 490, 600, "light fire", game.light_fire_progress)
    outputs.primitives << button(       :save_game, 490, 540, "save")
    outputs.primitives << button(       :load_game, 490, 480, "load")
    outputs.primitives << button(      :reset_game, 490, 420, "reset")
    outputs.labels << [640, 30, "the fire is #{game.fire}", 0, 1]
  end

  def tick
    state.layout.highlights ||= []
    game.tick if game.should_tick?
    render
    process_input
  end

  def button_clicked?
    return nil unless click_pos
    button = @button_list.find do |k, v|
      click_pos.inside_rect? v[:primitives].last.rect
    end
    return unless button
    method_to_call = "#{button[0]}_clicked".to_sym
    if self.respond_to? method_to_call
      border = button[1][:primitives].last

      state.layout.highlights << state.new_entity(:highlight) do |h|
        h.x = border.x
        h.y = border.y
        h.w = border.w
        h.h = border.h
        h.max_lifetime = 10
        h.lifetime = h.max_lifetime
        h.color = [120, 120, 180]
      end

      self.send method_to_call
    else
      border = button[1][:primitives].last

      state.layout.highlights << state.new_entity(:highlight) do |h|
        h.x = border.x
        h.y = border.y
        h.w = border.w
        h.h = border.h
        h.max_lifetime = 4
        h.lifetime = h.max_lifetime
        h.color = [120, 80, 80]
      end

      puts "It looks like #{method_to_call} doesn't exists on TextedBasedGamePresenter. Please add this method:"
      puts "Just copy the code below and put it in the #{TextedBasedGamePresenter} class definition."
      puts ""
      puts "```"
      puts "class TextedBasedGamePresenter <--- find this class and put the method below in it"
      puts ""
      puts "  def #{method_to_call}"
      puts "    puts 'Yay that worked!'"
      puts "  end"
      puts ""
      puts "end <-- make sure to put the #{method_to_call} method in between the `class` word and the final `end` statement."
      puts "```"
      puts ""
    end
  end

  def alert_dismiss_clicked
    game.clear_storyline
    game.go_to_room
  end

  def click_pos
    return nil unless inputs.mouse.click
    return inputs.mouse.click.point
  end

  def process_input
    button = button_clicked?
  end

  def game
    @game
  end

  def light_fire_clicked
    game.light_fire
  end

  def save_game_clicked
    game.save
  end

  def reset_game_clicked
    game.reset
  end

  def load_game_clicked
    game.load
  end
end

$text_based_rpg = TextedBasedGamePresenter.new

def tick args
  $text_based_rpg.state = args.state
  $text_based_rpg.outputs = args.outputs
  $text_based_rpg.inputs = args.inputs
  $text_based_rpg.tick
end
